# -*- coding: utf-8 -*-
# (c) 2021-2023, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import sys
import subprocess

from .base import AndeboxAction
from ..exceptions import AndeboxException


def _make_default_tox_ini():
    return f""";
; andebox tox-test's tox.ini -- this file is not overwritten by andebox
[tox]
isolated_build = true
envlist = ac211, ac212, ac213, ac214, ac215, a4, a5, a6, a7, a8, dev
skipsdist = true

[testenv]
passenv = PWD, HOME
skip_install = true
allowlist_externals = andebox
basepython = {sys.executable}
deps =
  andebox>=0.36
  ac211: ansible-core~=2.11.0
  ac212: ansible-core~=2.12.0
  ac213: ansible-core~=2.13.0
  ac214: ansible-core~=2.14.0
  ac215: ansible-core~=2.15.0
  a4: ansible~=4.0.0
  a5: ansible~=5.0.0
  a6: ansible~=6.0.0
  a7: ansible~=7.0.0
  a8: ansible~=8.0.0
  dev: https://github.com/ansible/ansible/archive/devel.tar.gz
commands = andebox test -- {{posargs}}
"""


class ToxTestError(AndeboxException):
    pass


class ToxTestAction(AndeboxAction):
    name = "tox-test"
    help = "runs ansible-test within tox, for testing in multiple ansible versions"
    args = [
        dict(names=("--env", "-e"),
             specs=dict(help="tox environments to run the test in")),
        dict(names=("--list", "-l"),
             specs=dict(action="store_true", help="list all tox environments (tox -a)")),
        dict(names=("--recreate", "-r"),
             specs=dict(action="store_true", help="force recreation of virtual environments (tox -r)")),
        dict(names=("ansible_test_params", ),
             specs=dict(nargs="*")),
    ]
    tox_ini_filename = ".andebox-tox-test.ini"

    @classmethod
    def make_parser(cls, subparser):
        action_parser = super(ToxTestAction, cls).make_parser(subparser)
        action_parser.epilog = "Notice the use of '--' to delimit andebox's options from tox's"
        action_parser.usage = "%(prog)s [-h] [--env ENV] -- [ansible_test_params ...]"

    def run(self, args):
        if not os.path.exists(self.tox_ini_filename):
            with open(self.tox_ini_filename, "w") as tox_ini:
                tox_ini.write(_make_default_tox_ini())

        cmd_args = ["tox", "-c", self.tox_ini_filename]
        if args.list:
            cmd_args.append("-a")
        if args.recreate:
            cmd_args.append("-r")
        if args.env:
            cmd_args.extend(["-e", args.env])
        cmd_args.append("--")
        cmd_args.extend(args.ansible_test_params)
        rc = subprocess.call(cmd_args)

        if rc != 0:
            raise ToxTestError("Error running tox (rc={0})".format(rc))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) 2021-2022, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

__version__ = '0.51'

import argparse
import sys
import signal

from .exceptions import AndeboxException
from .actions.ansibletest import AnsibleTestAction
from .actions.ignorefile import IgnoreLinesAction
from .actions.runtime import RuntimeAction
from .actions.toxtest import ToxTestAction
from .actions.vagrant import VagrantAction
from .actions.docsite import DocsiteAction


_actions = [AnsibleTestAction, IgnoreLinesAction, RuntimeAction, ToxTestAction, VagrantAction, DocsiteAction]


def _make_parser():
    parser = argparse.ArgumentParser(prog="andebox", description="Ansible Developer (Tool)Box v{}".format(__version__))
    parser.add_argument("--version",
                        action="version",
                        version="%(prog)s {0}".format(__version__))
    parser.add_argument("--collection", "-c",
                        help="fully qualified collection name (not necessary if a proper galaxy.yml file is available)")
    subparser = parser.add_subparsers(dest="action", required=True)

    for action in _actions:
        action.make_parser(subparser)

    return parser


class AndeBox:
    actions = {}
    parser = None

    def __init__(self) -> None:
        self.add_actions(*_actions)

    def add_actions(self, *actions):
        self.actions.update({ac.name: ac() for ac in actions})

    def build_argparser(self):
        self.parser = _make_parser()

    def run(self):
        self.build_argparser()
        args = self.parser.parse_args()
        action = self.actions[args.action]
        action.run(args)


def run():
    try:
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        box = AndeBox()
        box.run()
        return 0
    except KeyboardInterrupt:
        print("Interrupted by user", file=sys.stderr)
        return 2
    except (AndeboxException, BrokenPipeError) as e:
        print(str(e), file=sys.stderr)
        return 1



# -*- coding: utf-8 -*-
# (c) 2024, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import subprocess
import webbrowser
from contextlib import contextmanager
from pathlib import Path

from .base import AndeboxAction
from ..exceptions import AndeboxException
from ..context import ansible_collection_tree, determine_collection, binary_path


@contextmanager
def set_dir(path):
    previous = Path().absolute()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(previous)


class DocsiteAction(AndeboxAction):
    name = "docsite"
    help = "builds collection docsite"
    args = [
        dict(names=("--keep", "-k"),
             specs=dict(action="store_true", help="Keep temporary collection directory after execution")),
        dict(names=("--venv", "-V"),
             specs=dict(help="""path to the virtual environment where andebox and ansible are installed""")),
        dict(names=("--open", "-o"),
             specs=dict(action="store_true", help="Open browser pointing to main page after build")),
        dict(names=("--dest-dir", "-d"),
             specs=dict(help="Directory which should contain the docsite", default=".builtdocs")),
    ]

    def run(self, args):
        try:
            namespace, collection = determine_collection(args.collection)

            with ansible_collection_tree(namespace, collection, args.keep) as collection_dir:
                os.makedirs(args.dest_dir, mode=0o755, exist_ok=True)
                if not os.path.exists(os.path.join(args.dest_dir, "build.sh")):
                    subprocess.run([
                            binary_path(args.venv, "antsibull-docs"),
                            "sphinx-init", "--use-current", "--lenient", f"{namespace}.{collection}", "--dest-dir", args.dest_dir
                        ],
                        cwd=collection_dir,
                        check=True
                    )

                with set_dir(args.dest_dir):
                    subprocess.run([binary_path(args.venv, "python"), "-m", "pip", "install", "-qr", "requirements.txt"], check=True)
                    subprocess.run(["./build.sh"], check=True)
        except Exception as e:
            raise AndeboxException("Error running when building docsite") from e

        if args.open:
            webbrowser.open(f"{os.path.join(args.dest_dir, 'build', 'html', 'index.html')}")



# -*- coding: utf-8 -*-
# (c) 2023, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import shutil
import sys
import tempfile
from contextlib import contextmanager

import yaml

coll_copy_exclusion = ('.git', '.tox', '.venv', '.virtualvenv', 'venv', 'virtualenv')


def _copy_collection(coll_dir):
    # copy files to tmp ansible coll dir
    with os.scandir() as it:
        for entry in it:
            if any(entry.name.startswith(x) for x in coll_copy_exclusion):
                continue
            if entry.is_dir():
                shutil.copytree(entry.name, os.path.join(coll_dir, entry.name), symlinks=True, ignore_dangling_symlinks=True)
            else:
                shutil.copy(entry.name, os.path.join(coll_dir, entry.name), follow_symlinks=False)


@contextmanager
def ansible_collection_tree(namespace, collection, keep=False):
    top_dir = ""
    coll_dir = ""
    try:
        top_dir = tempfile.mkdtemp(prefix="andebox.")
        coll_dir = os.path.join(top_dir, "ansible_collections", namespace, collection)
        os.makedirs(coll_dir)
        print(f"collection = {namespace}.{collection}", file=sys.stderr)
        print(f"directory  = {coll_dir}", file=sys.stderr)

        _copy_collection(coll_dir)
        os.putenv('ANSIBLE_COLLECTIONS_PATH', ':'.join([top_dir] + os.environ.get('ANSIBLE_COLLECTIONS_PATH', '').split(':')))
        yield coll_dir

    finally:
        if keep:
            print('Keeping temporary directory: {0}'.format(coll_dir))
        else:
            print('Removing temporary directory: {0}'.format(coll_dir))
            shutil.rmtree(top_dir)


def read_coll_meta():
    with open("galaxy.yml") as galaxy_meta:
        meta = yaml.safe_load(galaxy_meta)
    return meta['namespace'], meta['name'], meta['version']


def determine_collection(coll_arg):
    if coll_arg:
        coll_split = coll_arg.split('.')
        return '.'.join(coll_split[:-1]), coll_split[-1]
    return read_coll_meta()[:2]


def binary_path(venv, binary):
    _list = ([venv, "bin"] if venv else []) + [binary]
    return os.path.join(*_list)


def copy_exclude_lines(src, dest, exclusion_filenames):
    with open(src, "r") as src_file, open(dest, "w") as dest_file:
        for line in src_file.readlines():
            if not any(line.startswith(f) for f in exclusion_filenames):
                dest_file.write(line)



# -*- coding: utf-8 -*-
# (c) 2021, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import re
import sys
from functools import reduce
from looseversion import LooseVersion

from .base import AndeboxAction


class IgnoreFileEntry:
    pattern = re.compile(r'^(?P<filename>\S+)\s(?P<ignore>\S+)(?:\s+#\s*(?P<comment>\S.*\S))?\s*$')
    filter_files = None
    filter_checks = None
    file_parts_depth = None

    def __init__(self, filename, ignore, comment):
        self.filename = filename
        self._file_parts = self.filename.split("/")

        if ':' in ignore:
            self.ignore, self.error_code = ignore.split(":")
        else:
            self.ignore, self.error_code = ignore, None
        self.comment = comment

    @property
    def ignore_check(self):
        return "{0}:{1}".format(self.ignore, self.error_code) if self.error_code else self.ignore

    @property
    def rebuilt_comment(self):
        return " # {0}".format(self.comment) if self.comment else ""

    @property
    def file_parts(self):
        if self.file_parts_depth is None:
            return os.path.join(*self._file_parts)

        return os.path.join(*self._file_parts[:self.file_parts_depth])

    def __str__(self):
        return "<IgnoreFileEntry: {0} {1}{2}>".format(self.filename, self.ignore_check, self.rebuilt_comment)

    def __repr__(self):
        return str(self)

    @staticmethod
    def parse(line):
        match = IgnoreFileEntry.pattern.match(line)
        if not match:
            raise ValueError("Line cannot be parsed as an ignore-file entry: {0}".format(line))

        ffilter = IgnoreFileEntry.filter_files
        if ffilter is not None:
            ffilter = ffilter if isinstance(ffilter, re.Pattern) else re.compile(ffilter)
            if not ffilter.search(match.group("filename")):
                return None

        ifilter = IgnoreFileEntry.filter_checks
        if ifilter is not None:
            ifilter = ifilter if isinstance(ifilter, re.Pattern) else re.compile(ifilter)
            if not ifilter.search(match.group("ignore")):
                return None

        return IgnoreFileEntry(match.group("filename"), match.group("ignore"), match.group("comment"))


class ResultLine:
    def __init__(self, file_part, ignore_check, count=1):
        self.file_part = file_part
        self.ignore_check = ignore_check
        self.count = count

    def increase(self):
        self.count = self.count + 1
        return self

    def __lt__(self, other):
        return self.count < other.count

    def __le__(self, other):
        return self.count <= other.count

    def __gt__(self, other):
        return self.count > other.count

    def __ge__(self, other):
        return self.count >= other.count

    def __eq__(self, other):
        return self.count == other.count

    def __ne__(self, other):
        return self.count != other.count

    def __str__(self):
        r = ["{0:6} ".format(self.count)]
        if self.file_part:
            r.append(" ")
            r.append(self.file_part)
        if self.ignore_check:
            r.append(" ")
            r.append(self.ignore_check)
        return "".join(r)

    def __repr__(self):
        r = ["<ResultLine: ", str(self.count), ","]
        if self.file_part:
            r.append(" ")
            r.append(self.file_part)
        if self.ignore_check:
            r.append(" ")
            r.append(self.ignore_check)
        r.append(">")
        return "".join(r)


_ignore_path = os.path.join('.', 'tests', 'sanity')
try:
    with os.scandir(os.path.join(_ignore_path)) as sanity_dir:
        _ignore_versions = sorted([
            str(LooseVersion(entry.name[7:-4]))
            for entry in sanity_dir
            if entry.name.startswith("ignore-") and entry.name.endswith(".txt")
        ])
except FileNotFoundError:
    _ignore_versions = []


class IgnoreLinesAction(AndeboxAction):
    name = "ignores"
    help = "gathers stats on ignore*.txt file(s)"
    args = [
        dict(names=["--ignore-file-spec", "-ifs"],
             specs=dict(choices=_ignore_versions + ["-"],
                        help=(
                            "Use the ignore file matching this Ansible version. "
                            "The special value '-' may be specified to read "
                            "from stdin instead. If not specified, will use all available files. "
                            "If no choices are presented, the collection structure was not recognized."))),
        dict(names=["--depth", "-d"],
             specs=dict(type=int,
                        help="Path depth for grouping files")),
        dict(names=["--filter-files", "-ff"],
             specs=dict(type=re.compile,
                        help="Regexp matching file names to be included")),
        dict(names=["--filter-checks", "-fc"],
             specs=dict(type=re.compile,
                        help="Regexp matching checks in ignore files to be included")),
        dict(names=["--suppress-files", "-sf"],
             specs=dict(action="store_true",
                        help="Supress file names from the output, consolidating the results")),
        dict(names=["--suppress-checks", "-sc"],
             specs=dict(action="store_true",
                        help="Suppress the checks from the output, consolidating the results")),
        dict(names=["--head", "-H"],
             specs=dict(type=int,
                        default=10,
                        help=(
                            "Number of lines to display in the output: leading lines if "
                            "positive, trailing lines if negative, all lines if zero."))),
    ]

    # pylint: disable=consider-using-with
    def make_fh_list_for_version(self, version):
        if version == "-":
            return [sys.stdin]
        if version:
            return [open(os.path.join(_ignore_path, 'ignore-{0}.txt'.format(version)))]

        with os.scandir(os.path.join(_ignore_path)) as it:
            return [open(os.path.join(_ignore_path, entry.name))
                    for entry in it
                    if entry.name.startswith("ignore-") and entry.name.endswith(".txt")]

    @staticmethod
    def read_ignore_file(fh):
        result = []
        with fh:
            for line in fh.readlines():
                entry = IgnoreFileEntry.parse(line)
                if entry:
                    result.append(entry)
        return result

    def retrieve_ignore_entries(self, version):
        return reduce(lambda a, b: a + b,
                      [self.read_ignore_file(fh) for fh in self.make_fh_list_for_version(version)])

    @staticmethod
    def filter_lines(lines, num):
        if num == 0:
            return lines
        return lines[num:] if num < 0 else lines[:num]

    def run(self, args):
        if args.filter_files:
            IgnoreFileEntry.filter_files = args.filter_files
        if args.filter_checks:
            IgnoreFileEntry.filter_checks = args.filter_checks
        if args.depth:
            IgnoreFileEntry.file_parts_depth = args.depth

        try:
            ignore_entries = self.retrieve_ignore_entries(args.ignore_file_spec)
        except Exception as e:
            print("Error reading ignore file {0}: {1}".format(args.ignore_file_spec, str(e)), file=sys.stderr)
            raise e

        count_map = {}
        for entry in ignore_entries:
            fp = entry.file_parts if not args.suppress_files else ""
            ic = entry.ignore_check if not args.suppress_checks else ""
            key = fp + "|" + ic
            count_map[key] = count_map.get(key, ResultLine(fp, ic, 0)).increase()

        lines = [str(s) for s in sorted(count_map.values(), reverse=True)]
        print("\n".join(self.filter_lines(lines, args.head)))



# -*- coding: utf-8 -*-
# (c) 2021, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import argparse
import os
import re
from functools import partial

import yaml

from .base import AndeboxAction
from ..context import read_coll_meta


PLUGIN_TYPES = ('connection', 'lookup', 'modules', 'doc_fragments', 'module_utils', 'callback', 'inventory')


def info_type(types, v):
    try:
        r = [t for t in types if t.startswith(v.lower())]
        return r[0][0].upper()
    except IndexError as e:
        raise argparse.ArgumentTypeError("invalid value: {0}".format(v)) from e


class RuntimeAction(AndeboxAction):
    RUNTIME_TYPES = ('redirect', 'tombstone', 'deprecation')
    name = "runtime"
    help = "returns information from runtime.yml"
    args = [
        dict(names=["--plugin-type", "-pt"],
             specs=dict(choices=PLUGIN_TYPES,
                        help="Specify the plugin type to be searched")),
        dict(names=["--regex", "--regexp", "-r"],
             specs=dict(action="store_true",
                        help="Treat plugin names as regular expressions")),
        dict(names=["--info-type", "-it"],
             specs=dict(type=partial(info_type, RUNTIME_TYPES),
                        help="Restrict type of response elements. Must be in {0}, may be shortened "
                             "down to one letter.".format(RUNTIME_TYPES))),
        dict(names=["plugin_names"],
             specs=dict(nargs='+')),
    ]
    name_tests = []
    current_version = None
    info_type = None

    def print_runtime(self, name, node):
        def is_info_type(_type):
            return self.info_type is None or self.info_type.lower() == _type.lower()

        redir, tomb, depre = [node.get(x) for x in self.RUNTIME_TYPES]
        if redir and is_info_type('R'):
            print('R {0}: redirected to {1}'.format(name, redir))
        elif tomb and is_info_type('T'):
            print('T {0}: terminated in {1}: {2}'.format(name, tomb['removal_version'], tomb['warning_text']))
        elif depre and is_info_type('D'):
            print('D {0}: deprecation in {1} (current={2}): {3}'.format(
                  name, depre['removal_version'], self.current_version, depre['warning_text']))

    def runtime_process_plugin(self, plugin_routing, plugin_types):
        for plugin_type in plugin_types:
            matching = [
                name
                for name in plugin_routing[plugin_type]
                if any(test(name) for test in self.name_tests)
            ]
            for name in matching:
                self.print_runtime('{0} {1}'.format(plugin_type, name), plugin_routing[plugin_type][name])

    def run(self, args):
        with open(os.path.join("meta", "runtime.yml")) as runtime_yml:
            runtime = yaml.safe_load(runtime_yml)

        plugin_types = [args.plugin_type] if args.plugin_type else PLUGIN_TYPES
        _, _, self.current_version = read_coll_meta()
        self.info_type = args.info_type

        def name_test(name, other):
            if name.endswith('.py'):
                name = name.split('/')[-1]
                name = name.split('.')[0]
            return name == other

        test_func = re.search if args.regex else name_test
        self.name_tests = [partial(test_func, n) for n in args.plugin_names]

        self.runtime_process_plugin(runtime['plugin_routing'], plugin_types)
